/*
 * For licensing see accompanying LICENSE file.
 *
 * Copyright (C) 2022 Apple Inc. All Rights Reserved.
 */

import { csv, json, synth, vega, vegaCompressed } from './loaders';
import App from './components/App.svelte';
import type { Confusion } from './confusions';
import { defaults } from './specification';
import type { Spec } from './specification';

type Loader = () => Promise<Array<Confusion>>;

type Example = {
    description: string;
    loader: Loader;
    spec: Spec;
}

const examples: Array<Example> = [
    {
        description: 'Fruit',
        loader: () => json('data/fruit.json'),
        spec: {
            ...defaults,
            classes: ['fruit'],
            normalization: 'total',
        },
    },
    {
        description: 'Beverages',
        loader: () => vega('data/beverages.json'),
        spec: {
            ...defaults,
            classes: ['beverage'],
            normalization: 'total',
        },
    },
    {
        description: 'Synthetic',
        loader: async () => synth(),
        spec: {
            ...defaults,
            classes: ['vehicle'],
            normalization: 'total',
        },
    },
    {
        description: 'Synthetic (nested, filtered)',
        loader: async () => synth(),
        spec: {
            ...defaults,
            classes: ['vehicle', 'state'],
            filter: ['vehicle:motorized'],
            normalization: 'total',
        },
    },
    {
        description: 'Synthetic (conditioning)',
        loader: async () => synth(),
        spec: {
            ...defaults,
            classes: ['vehicle', 'animal'],
            where: { qualifier: 'actual', label: 'state', is: 'state:moving' },
            normalization: 'total',
        },
    },
    {
        description: 'Case Study 1 (App Prediction)',
        loader: () => csv('data/chi-paper/case-study-1.csv'),
        spec: {
            ...defaults,
            classes: ['root'],
            measures: ['accuracy', 'precision', 'recall'],
            encoding: 'color',
            normalization: 'total',
        },
    },
    {
        description: 'Case Study 2 (ImageNet)',
        loader: () => vegaCompressed('data/chi-paper/case-study-2.json'),
        spec: {
            ...defaults,
            classes: ['root'],
            measures: ['precision', 'recall', 'countObserved'],
            encoding: 'color',
            normalization: 'total',
            collapsed: ['', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:weapon-arm-weap:gun', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:shade', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:lepidopterous i:butterfly:danaid-danaid b', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:housing-lodging:dwelling-home-d', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:receptacle', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:measure', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:mastiff', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:spiny-finned fi:percoid fish-pe', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:blind-screen:window blind', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:floor cover-flo:rug-carpet-carp', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr:meter', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:monitor-monitor', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:ganoid-ganoid f', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:edible fruit:apple:eating apple-de', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:plant-flora-pla:vascular plant-:spermatophyte-p:angiosperm-flow:flower:orchid-orchidac', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:box', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:sporting dog-gu:pointer-Spanish', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:landing-landing', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:seat:sofa-couch-loun:convertible-sof', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:game equipment:game', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:article:ware:tableware', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment:institution:penal instituti', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:gallinule-marsh', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :ingredient-fixi:flavorer-flavou:condiment:sauce:spaghetti sauce', 'root:entity:physical entity:matter:substance:food-nutrient:fare', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :ingredient-fixi:flavorer-flavou:condiment:dip', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:optical device:lens-lense-lens:converging lens', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:bulldog-English', 'root:entity:abstraction-abs:communication:signal-signalin:visual signal:light', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:weapon-arm-weap:gun:firearm-piece-s:autoloader-self', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:wagon-waggon:cart', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:writing impleme:pen', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs:barrier:movable barrier', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:homopterous ins', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:cutting impleme:cutter-cutlery-:edge tool:knife', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:rail:coot', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:attire-garb-dre', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:roof', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:area', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:apparel-wearing:workwear', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:screen-cover-co', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:crocodilian rep:alligator-gator', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:cuculiform bird', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr:measuring stick', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:footwear', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:footwear:hosiery-hose', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:support:structural memb:upright-vertica:post', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:edible fruit:berry', 'root:entity:physical entity:matter:substance:food-nutrient:beverage-drink-:coffee-java', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:shorebird-shore', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:sporting dog-gu:spaniel:springer spanie', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument:pipe', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:fungus:basidiomycete-b', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:mechanical devi:ride', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:scientific inst:magnifier:microscope:light microscop:hand glass-simp', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:boat', 'root:entity:physical entity:object-physical:part-portion:piece:piece of cloth-:towel', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:viper', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:aircraft:heavier-than-ai:airplane-aeropl', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:mechanical devi:winder', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:public transpor:train-railroad ', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:pinniped mammal', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:weapon-arm-weap:gun:firearm-piece-s:autoloader-self:automatic firea:automatic rifle', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:headdress-headg:cap', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:creation:product-product', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:cleaning implem', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:bird of prey-ra', 'root:entity:physical entity:object-physical:whole-unit', 'root:entity:physical entity:matter:solid', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:squash', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:hand tool', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe:oscine-oscine b:finch:bunting', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:system:communication s:telecommunicati', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:drill', 'root:entity:physical entity:object-physical:geological form:natural elevati:ridge:reef', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:bottle:jug', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:rodent-gnawer', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu:peer-equal-matc:associate:participant', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:padding-cushion:cushion', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:elapid-elapid s:cobra', 'root:entity:abstraction-abs', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:amphibian:salamander', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:handwear-hand w', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:plant-flora-pla:vascular plant-:spermatophyte-p:angiosperm-flow:flower', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:monkey:Old World monke', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr:timepiece-timek:sandglass', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:shirt', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods:durables-durabl:appliance:home appliance-:white goods', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:fastener-fasten', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:wild sheep:mountain sheep', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:percussion inst', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument:woodwind-woodwi:beating-reed in', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:crustacean', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:strengthener-re:brace', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:swine', 'root:entity:abstraction-abs:communication:sign', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:cabinet', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:bag', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:article', 'root:entity:physical entity:matter:substance:agent:drug:drug of abuse-s', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:lagomorph-gnawi:leporid-leporid:rabbit-coney-co', 'root:entity:physical entity:object-physical:geological form:natural depress', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:coraciiform bir', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:archosaur-archo', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe:oscine-oscine b:titmouse-tit', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :ingredient-fixi', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:shepherd dog-sh:Belgian sheepdo', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:dish:stew', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:stringed instru:piano-pianofort', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:headdress-headg:hat-chapeau-lid', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:terrier:wirehair-wireha', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:swan', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment:place of busine', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:cetacean-cetace:whale:toothed whale:dolphin', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:optical device:lens-lense-lens', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:poodle-poodle d', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:seed:oilseed-oil-ric', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:public transpor:train-railroad :passenger train', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:pachyderm', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:toiletry-toilet:cosmetic', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:footwear-footge:shoe', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:course:entree-main cou', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:mechanical devi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:sports equipmen', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:cartilaginous f:elasmobranch-se:shark:requiem shark', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:constrictor:boa:python', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:crustacean:decapod crustac', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:seed:grain-caryopsis', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:sporting dog-gu:retriever', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:article:ware:tableware:cutlery-eating ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :armored vehicle', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:antelope', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:lamp', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:squash:summer squash', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:supporting stru', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:support:structural memb', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :starches:bread-breadstuf:loaf of bread-l', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:bicycle-bike-wh', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:colubrid snake-', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:fastener-fasten:clip', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:control-control', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:goat-caprine an', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:mechanical devi:machine-simple ', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:elapid-elapid s', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:course:dessert-sweet-a:pudding-pud', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:terrier:bullterrier-bul', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:computer-comput:digital compute:personal comput', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:optical instrum:spectacles-spec', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:creation:product-product:work-piece of w:publication', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:cetacean-cetace', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:seabird-sea bir:sphenisciform s:penguin', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:archosaur-archo:dinosaur', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:cloth covering:dressing-medica:bandage-patch:adhesive bandag', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:supporting stru:support', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:ape:anthropoid ape', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods:durables-durabl:appliance:home appliance-', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:constrictor', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:outerwear-overc', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:sports equipmen:weight-free wei', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:cartilaginous f', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:mollusk-mollusc:cephalopod-ceph', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:agamid-agamid l', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:gallinaceous bi:game bird:phasianid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:connection-conn:attachment-bond:ligament', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:womans clothing', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:gear-parapherna', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:spiny-finned fi:percoid fish-pe:butterfly fish', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:litter', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :locomotive-engi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:lifting device', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe:oscine-oscine b:corvine bird', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:column-pillar', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:overgarment-out:coat', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:sea cow-sirenia', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:source of illum:lamp', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:plant-flora-pla:vascular plant-', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr:timepiece-timek:timer', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:boat:small boat', 'root:entity:physical entity:object-physical:geological form:natural elevati:ridge', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:lagomorph-gnawi:leporid-leporid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:seat:sofa-couch-loun', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:crustacean:decapod crustac:lobster:true lobster', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:cutting impleme:cutter-cutlery-:edge tool:ax-axe', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:housing-lodging', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:eraser', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:computer-comput:analog computer', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :starches:bread-breadstuf:bun-roll', 'root:entity:physical entity:matter', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:breathing devic', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device', 'root:entity:physical entity:matter:substance:food-nutrient', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:tank-storage ta:reservoir', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:electronic equi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:dish', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:soft-finned fis', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:blower', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:lemur', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:wall unit', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:strengthener-re', 'root:entity:physical entity', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:shelter:tent-collapsibl', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:crustacean:malacostracan c', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:bovine', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:apparatus-setup', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:shield:plate-scale-she:armor plate-arm:helmet', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:terrier:schnauzer', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:shorebird-shore:sandpiper', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:house:residence', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:trap:web-entanglemen', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:parrot:cockatoo', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:lepidopterous i', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:cartilaginous f:elasmobranch-se:shark', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs:barrier', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :motor vehicle-a:truck-motortruc:van', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :concoction-mixt', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:shelter:canopy', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:monkey', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :motor vehicle-a:motorcycle-bike:minibike-motorb', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:monkey:New World monke', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:handwear-hand w:glove', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:gallinaceous bi:game bird:phasianid:pheasant:peafowl-bird of', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:seabird-sea bir:pelagic bird-oc', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:photographic eq', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:hound-hound dog', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:game equipment:ball', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:arachnid-arachn', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:toy dog-toy:toy spaniel', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:neckwear', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:outerwear-overc:gown-robe', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:plant-flora-pla:vascular plant-:spermatophyte-p:angiosperm-flow:flower:orchid-orchidac:ladys slipper-l', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:place of worshi:shrine', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:soft-finned fis:salmonid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:blind-screen', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:sled dog-sledge', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:headdress-headg:helmet', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:cloth covering:dressing-medica:bandage-patch', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:seabird-sea bir', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods:durables-durabl:appliance:home appliance-:kitchen applian:coffee maker', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:game equipment:game:puzzle', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:sporting dog-gu:spaniel:water spaniel', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:goat-caprine an:wild goat', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:aircraft:lighter-than-ai', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:computer-comput:digital compute:personal comput:portable comput', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:feline-felid', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu:explorer-advent', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:terrier', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:crustacean:decapod crustac:crab', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument:woodwind-woodwi:beating-reed in:double-reed ins', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:mechanical devi:machine-simple :wheel', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:plant-flora-pla:vascular plant-:spermatophyte-p', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :starches', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:amphibian:salamander:newt-triton', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:case', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:food fish', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:supporting stru:framework:rack', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:navigational in:compass', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:hymenopterous i', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument:woodwind-woodwi:beating-reed in:single-reed ins', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :motor vehicle-a:car-auto-automo', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment:institution:penal instituti:correctional in', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:overgarment-out:coat:raincoat-waterp', 'root:entity', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:seed', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:place of worshi', 'root:entity:physical entity:object-physical:whole-unit:natural object', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:venomous lizard', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:lepidopterous i:butterfly:nymphalid-nymph', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:creation:product-product:work-piece of w', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod', 'root:entity:physical entity:object-physical:part-portion:piece', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:ship:passenger ship', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:heater-warmer', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:constrictor:boa', 'root:entity:abstraction-abs:communication:signal-signalin', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:sailing vessel-', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:source of illum:lamp:lantern', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:screen', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:cutting impleme:cutter-cutlery-', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:spitz', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:feline-felid:cat-true cat', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:metatherian:marsupial-pouch:kangaroo', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:creation', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:table', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:skirt', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:applicator-appl', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:feline-felid:cat-true cat:domestic cat-ho', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:house:residence:religious resid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:seat', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:Old World buffa', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:shaker', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:cruciferous veg', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:support', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:utensil:kitchen utensil:cooking utensil:pan-cooking pan', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:rocket-projecti', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:cloth covering', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:stick:staff', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:electronic equi:audio system-so:reproducer:stereo-stereoph', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:scientific inst:magnifier:telescope-scope:astronomical te', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu:contestant', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:reflector', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:scientific inst:magnifier:telescope-scope', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:elapid-elapid s:mamba:black mamba-Den', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:terrier:wirehair-wireha:Welsh terrier', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe:oscine-oscine b', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:archosaur-archo:dinosaur:ornithischian-o', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:scientific inst:magnifier', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:garden tool-law', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:bar:lever', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu:peer-equal-matc', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:ship', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:parrot:lory', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:medical instrum', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:waterfowl-water:anseriform bird:duck', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:sweater-jumper', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:ship:cargo ship-carg', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:theater-theatre', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:sailing vessel-:sailboat-sailin', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:apparatus-setup:duplicator-copi', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:coelenterate-cn:anthozoan-actin', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:ship:warship-war ves', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:protective garm', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:drinking vessel:mug', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:gecko', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:acoustic device', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:mollusk-mollusc:gastropod-univa', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:filter', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs:barrier:fence-fencing:rail fence', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:camel', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:bar:lever:key', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:scientific inst', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:support:rack-stand', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:cap', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:bridge-span', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:gear-parapherna:rig', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:fox', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:public transpor', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:top-cover', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:basin', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:power tool:power saw-saw-s', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:solanaceous veg', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:arachnid-arachn:acarine', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:terrier:fox terrier', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:gallinaceous bi:game bird:grouse', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe:oscine-oscine b:finch', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:sporting dog-gu:setter', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:bin', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:stringed instru', 'root:entity:physical entity:matter:substance', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:decoration-orna:adornment:jewelry-jewelle', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:surface', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:basket-handbask', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:car-railcar-rai', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:watchdog-guard ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment:place of busine:mercantile esta:shop-store', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:dish:snack food', 'root:entity:physical entity:object-physical:geological form:shore', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:hound-hound dog:foxhound', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:stick', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu:peer-equal-matc:associate', 'root:entity:physical entity:matter:substance:material-stuff', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:gear-parapherna:kit-outfit', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:system', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:handcart-pushca', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:lacertid lizard', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:aircraft', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:seat:bench', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:utensil:kitchen utensil:cooking utensil', 'root:entity:abstraction-abs:attribute:shape-form', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:toiletry-toilet:cosmetic:makeup-make-up-', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:feline-felid:big cat-cat', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:waterfowl-water:anseriform bird:duck:sea duck:merganser-fish ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:swimsuit-swimwe', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:sports equipmen:basketball equi', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:iguanid-iguanid', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:pinniped mammal:seal:eared seal', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:footwear:hosiery-hose:tights-leotards', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:crustacean:decapod crustac:lobster', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:heron', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:electrical devi', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:gallinule-marsh:purple gallinul', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:toiletry-toilet', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:cartilaginous f:elasmobranch-se:shark:mackerel shark', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:dish:soup', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:game equipment', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:ape:anthropoid ape:great ape-pongi', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:crocodilian rep', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:dish:bowl', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:viper:pit viper:rattlesnake-rat', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:jar', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:photographic eq:camera-photogra', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:farm machine', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:drinking vessel', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:scientific inst:magnifier:microscope:light microscop', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:rail', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:procyonid', 'root:entity:physical entity:matter:substance:agent:drug:drug of abuse-s:alcohol-alcohol:mixed drink:punch', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:musteline mamma', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:spiny-finned fi:scorpaenoid-sco:scorpaenid-scor', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:course:dessert-sweet-a:frozen dessert', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:tank-storage ta', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:plant-flora-pla:vascular plant-:spermatophyte-p:angiosperm-flow', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:arch', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:stringed instru:guitar', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:sports equipmen:gymnastic appar', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:armor-armour:body armor-body', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:hound-hound dog:greyhound', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:dish:snack food:sandwich', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:bedroom furnitu:bed', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs:barrier:movable barrier:door', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument:woodwind-woodwi', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:hound-hound dog:wolfhound', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:house', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:footwear-footge', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:sports equipmen:baseball equipm', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:pachyderm:elephant', 'root:entity:abstraction-abs:communication:signal-signalin:visual signal', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:utensil:kitchen utensil', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:utensil:kitchen utensil:cooking utensil:cooker', 'root:entity:physical entity:matter:solid:food-solid food', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:brake', 'root:entity:abstraction-abs:attribute:shape-form:round shape:sphere:ball-globe-orb:globule', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:sheep', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:beetle:lamellicorn bee:scarabaeid beet', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:viverrine-viver', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:cruciferous veg:cabbage-chou', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :motor vehicle-a:truck-motortruc:van:passenger van', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:cutting impleme', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:sporting dog-gu', 'root:entity:physical entity:matter:substance:agent:drug', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:shorebird-shore:snipe', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:surface:horizontal surf:platform', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:computer-comput', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :starches:bread-breadstuf:cracker', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:weapon-arm-weap', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:rodent-gnawer:squirrel', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:mask:face mask', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul', 'root:entity:physical entity:matter:substance:agent:drug:drug of abuse-s:alcohol-alcohol:wine-vino', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:anguid lizard', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:wolf', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:waterfowl-water:anseriform bird:duck:sea duck', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu:explorer-advent:diver-frogman-u', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:gallinaceous bi:game bird', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:stringed instru:bowed stringed ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:savings bank-co', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:textile machine', 'root:entity:physical entity:matter:substance:agent:drug:drug of abuse-s:alcohol-alcohol', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:defensive struc', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:rodent-gnawer:squirrel:tree squirrel', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:excavation:pool:cistern', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:beetle:lamellicorn bee', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :motor vehicle-a:truck-motortruc', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:instrument of e', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:odonate', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:chest of drawer', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:odd-toed ungula', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra', 'root:entity:physical entity:matter:substance:material-stuff:paper', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:cetacean-cetace:whale:baleen whale-wh', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :starches:bread-breadstuf', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:shield', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:mechanical devi:pump', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum', 'root:entity:physical entity:object-physical:part-portion', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:wrapping-wrap-w:jacket', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:printer-printin', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:anapsid-anapsid:chelonian-chelo', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:memory device-s:magnetic disk-m', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:metatherian:marsupial-pouch:phalanger-oposs', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:aircraft:heavier-than-ai', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:sports equipmen:golf equipment', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods:durables-durabl', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment:place of busine:mercantile esta', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:trap', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:seabird-sea bir:pelecaniform se', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:neckwear:necktie-tie', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:dispenser', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:anapsid-anapsid:chelonian-chelo:turtle', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:archosaur-archo:dinosaur:ornithischian-o:ceratopsian-hor', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:utensil:kitchen utensil:cooking utensil:turner-food tur', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:tower', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:glass-drinking ', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:heron:egret', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:scarf', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:amphibian:salamander:ambystomid-amby', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:metatherian', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods:durables-durabl:appliance:home appliance-:kitchen applian', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:rotating mechan:circle-round', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:footwear-footge:boot', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs:barrier:fence-fencing', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:root vegetable:potato-white po', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:phasmid-phasmid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:fastener-fasten:lock', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:soft-finned fis:cypriniform fis:cyprinid-cyprin', 'root:entity:physical entity:object-physical:geological form:natural elevati:mountain-mount', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:electronic equi:telephone-phone', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:reflector:mirror', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:public transpor:bus-autobus-coa', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:sports implemen', 'root:entity:physical entity:matter:substance:food-nutrient:feed-provender:fodder', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:footwear:hosiery-hose:stocking', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:plaything-toy', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:hand tool:opener:bottle opener', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:military vehicl', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:supporting stru:framework', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:wagon-waggon', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel:bottle', 'root:entity:physical entity:object-physical:geological form:natural elevati', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:gallinaceous bi:game bird:phasianid:pheasant', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:amphibian:frog-toad-toad ', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:edible fruit:apple', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:system:communication s', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:cloth covering:bedclothes-bed ', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:utensil', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:odd-toed ungula:equine-equid', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:spiny-finned fi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods:durables-durabl:appliance:home appliance-:kitchen applian:oven', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods:durables-durabl:appliance', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:anapsid-anapsid:chelonian-chelo:turtle:sea turtle-mari', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:box:strongbox-deedb', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa', 'root', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:package-parcel', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:electronic equi:audio system-so:reproducer', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:bedroom furnitu', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:coelenterate-cn:anthozoan-actin:coral', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:shield:plate-scale-she', 'root:entity:abstraction-abs:attribute:shape-form:round shape', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:boat:motorboat-power', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:cloth covering:dressing-medica', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:can-tin-tin can', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:source of illum:light-light sou', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:slot machine-co', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:creation:product-product:work-piece of w:publication:magazine-mag', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:decoration-orna:adornment', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:metatherian:marsupial-pouch', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:mollusk-mollusc', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:overgarment-out:cloak', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:spiny-finned fi:scorpaenoid-sco', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:bird of prey-ra:hawk', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:coelenterate-cn', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:dipterous insec', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:navigational in', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:coelenterate-cn:anthozoan-actin:coral:stony coral-mad', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:sweater-jumper:pullover-slipov', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:solanaceous veg:pepper', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:floor cover-flo', 'root:entity:physical entity:matter:substance:agent:drug:drug of abuse-s:alcohol-alcohol:mixed drink', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:surface:horizontal surf', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:neuropteron-neu', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:shepherd dog-sh', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:electronic devi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:armor-armour', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:ape:anthropoid ape:lesser ape', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:gallinaceous bi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building comple', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:shelter', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:weapon-arm-weap:gun:firearm-piece-s:pistol-handgun-', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:cetacean-cetace:whale', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:electrical devi:transducer:electro-acousti', 'root:entity:physical entity:matter:substance:food-nutrient:feed-provender', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:viper:pit viper', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:padding-cushion', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:wrapping-wrap-w', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:nightwear-sleep', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument', 'root:entity:abstraction-abs:communication', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:bear', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:utensil:kitchen utensil:cooking utensil:pot', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:toy dog-toy', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:blind-screen:curtain-drape-d', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building comple:plant-works-ind', 'root:entity:physical entity:object-physical', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment:place of busine:mercantile esta:marketplace-mar', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:calculator-calc', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:headdress-headg', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:connection-conn', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:edentate:sloth-tree slot', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:decoration-orna', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:rodent-gnawer:cavy', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:teiid lizard-te', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:griffon-Brussel', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs:barrier:movable barrier:gate', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :motor vehicle-a:motorcycle-bike', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:course:dessert-sweet-a', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:outbuilding:shed', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:sled-sledge-sle', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:shield:plate-scale-she:armor plate-arm', 'root:entity:physical entity:object-physical:geological form', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:rotating mechan:circle-round:disk-disc', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:arachnid-arachn:spider', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:Sennenhunde', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:spiny-finned fi:percoid fish-pe:damselfish-demo', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:snake-serpent-o:elapid-elapid s:mamba', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:brake:brake system-br:hydraulic brake', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:corgi-Welsh cor', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:power tool', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:oar', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr:timepiece-timek', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:undergarment-un', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:solanaceous veg:pepper:sweet pepper', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:container:vessel', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:bird of prey-ra:owl-bird of Min', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:electronic equi:telephone-phone:radiotelephone-', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:support:structural memb:upright-vertica', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:electronic equi:audio system-so', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:dictyopterous i', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:amphibian:frog-toad-toad :true frog-ranid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:obstruction-obs:barrier:grate-grating', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:outbuilding', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:prototherian:monotreme-egg-l', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:dish:snack food:sandwich:hamburger-beefb', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:wild dog', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :ingredient-fixi:flavorer-flavou', 'root:entity:physical entity:matter:substance:material-stuff:paper:tissue-tissue p', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building-edific:farm building', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:feline-felid:cat-true cat:wildcat', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:ratite-ratite b', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:hand tool:opener', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:baby bed-babys ', 'root:entity:physical entity:object-physical:geological form:spring-fountain', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:fastener-fasten:pin', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:person-individu:contestant:player-particip', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:apodiform bird', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:excavation:pool', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:root vegetable', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:writing impleme', 'root:entity:abstraction-abs:attribute:shape-form:round shape:sphere', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :tracked vehicle', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:seat:chair:chair of state', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:spacecraft-ball', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:fungus', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:optical device', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe:oscine-oscine b:thrush', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:rescue equipmen', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine:computer-comput:digital compute', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:mask', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:vest-waistcoat', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:beetle', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:bird of prey-ra:eagle-bird of J', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:weapon-arm-weap:gun:firearm-piece-s:autoloader-self:automatic firea', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate:ape', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:electronic devi:display-video d', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:connection-conn:attachment-bond', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:womans clothing:dress-frock', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:saurian:lizard:chameleon-chama', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:top-cover:cap', 'root:entity:abstraction-abs:attribute', 'root:entity:physical entity:matter:substance:food-nutrient:beverage-drink-', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:edible fruit:citrus-citrus f', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:sporting dog-gu:spaniel', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:diapsid-diapsid:crocodilian rep:crocodile', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr:timepiece-timek:watch-ticker', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :ingredient-fixi:flavorer-flavou:condiment:sauce', 'root:entity:physical entity:object-physical:part-portion:piece:piece of cloth-', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:defensive struc:fortification-m', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:stork', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:amphibian', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:office furnitur', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:source of illum', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:furnishing:furniture-piece:seat:chair', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:shelter', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:fan', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:hunting dog:hound-hound dog:coonhound', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument:brass-brass ins', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:seed:grain-caryopsis:kernel', 'root:entity:physical entity:object-physical:whole-unit:living thing-an', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:robe', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:primate', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:equipment:gear-parapherna:rig:drill rig-drill', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:bar', 'root:entity:physical entity:object-physical:geological form:natural elevati:ridge:bar', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:article:ware', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment:overgarment-out', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:boat:sea boat', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:keyboard', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:memory device-s', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:passerine-passe:oscine-oscine b:thrush:nightingale-Lus', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:weapon-arm-weap:gun:firearm-piece-s', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:machine', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g:vegetable-veggi:squash:winter squash', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:cartilaginous f:elasmobranch-se', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:wild sheep', 'root:entity:physical entity:matter:substance:food-nutrient:foodstuff-food :ingredient-fixi:flavorer-flavou:condiment', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:lepidopterous i:butterfly:pierid-pierid b', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:article:ware:tableware:cutlery-eating :spoon', 'root:entity:physical entity:matter:solid:food-solid food:produce-green g', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:rod', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:measuring instr:timepiece-timek:clock', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:signboard-sign', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:brake:brake system-br', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:cartilaginous f:elasmobranch-se:ray', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:craft:vessel-watercra:ship:warship-war ves:submersible-sub', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:optical instrum', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:soft-finned fis:cypriniform fis', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:orthopterous in', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:uniform', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:spiny-finned fi:plectognath-ple', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:seabird-sea bir:sphenisciform s', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:protective cove:sheath', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:aquatic vertebr:fish:bony fish:teleost fish-te:soft-finned fis:salmonid:salmon', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:self-propelled :motor vehicle-a', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:fabric-cloth-ma', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:mechanism:rotating mechan', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:plant-flora-pla', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:worm', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:echinoderm', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:electrical devi:transducer', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:course', 'root:entity:physical entity:matter:substance:food-nutrient:nutriment-nouri:dish', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:shorebird-shore:plover:turnstone', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:even-toed ungul:ruminant:bovid:bovine:cattle-cows-kin', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:garment', 'root:entity:physical entity:matter:substance:agent', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:toy dog-toy:toy spaniel:English toy spa', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:tool:cutting impleme:cutter-cutlery-:edge tool', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:reptile-reptili:anapsid-anapsid', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:runner', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:wading bird-wad:shorebird-shore:plover', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:edentate', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:implement:sharpener', 'root:entity:abstraction-abs:attribute:shape-form:round shape:sphere:ball-globe-orb', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:parrot', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:toiletry-toilet:cream-ointment-', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:waterfowl-water', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:pinniped mammal:seal', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:prototherian', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:carnivore:canine-canid:dog-domestic do:working dog:watchdog-guard :pinscher', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :invertebrate:arthropod:insect:lepidopterous i:butterfly', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:apparel-wearing', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:excavation', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:ungulate-hoofed:odd-toed ungula:equine-equid:horse-Equus cab', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:lagomorph-gnawi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:airfoil-aerofoi', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:floor cover-flo:mat', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:memorial-monume', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:restraint-const:safety belt-lif', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:keyboard instru', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:establishment:institution', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:aquatic bird:waterfowl-water:anseriform bird', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:wind instrument:free-reed instr', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:instrument:scientific inst:magnifier:microscope', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:mammal-mammalia:placental-place:aquatic mammal:cetacean-cetace:whale:toothed whale', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:commodity-trade:consumer goods', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:conveyance-tran:vehicle:wheeled vehicle:trailer-house t', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:piciform bird', 'root:entity:physical entity:object-physical:whole-unit:living thing-an:organism-being:animal-animate :chordate:vertebrate-cran:bird:cuculiform bird:cuckoo', 'root:entity:physical entity:object-physical:whole-unit:natural object:plant part-plan:plant organ:reproductive st:fruit:edible fruit', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:structure-const:building comple:plant-works-ind:factory-mill-ma', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:covering:clothing-articl:attire-garb-dre:hairpiece-false', 'root:entity:physical entity:object-physical:whole-unit:artifact-artefa:instrumentality:device:musical instrum:stringed instru:chordophone'],
        },
    },
    {
        description: 'Case Study 3 (Toxicity)',
        loader: () => json('data/chi-paper/case-study-3.json'),
        spec: {
            ...defaults,
            classes: ['toxic_mild'],
            measures: ['accuracy', 'countObserved', 'countActual', 'precision', 'recall', 'falsePositives', 'falseNegatives'],
            encoding: 'color',
            normalization: 'total',
        },
    },
];

new App({
    target: document.body,
    props: { examples },
});
